

// Cludged array constructor/destructor to support John.


#include <stdlib.h>
//#include <iostream.h>
extern "C" {
  int printf(char*, ...) ;
}
#include <stdarg.h>
#include "Seq.h"
#include "Procs.h"
#include "Range.h"
#include "Array.h"
#include "Cludge.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Cludge': Constructor and destructor for JHM-parametrised     //
//                 array.                                              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


Cludge :: Cludge(const int _typeLen, ProcessGroup* _procGrp, const int _rank,
                 const int data_rank, const int embed_rank,
                 int* vecs, void* _data) : Frame(_typeLen, _procGrp, _rank) {

  //printf("Cludge :: Cludge: _typeLen = %d _rank = %d data_rank = %d embed_rank = %d\n", _typeLen, _rank, data_rank, embed_rank) ;  // debug

  int* l   = vecs ;
  int* n   = l   + data_rank ;
  int* s   = n   + data_rank ;
  int* al  = s   + data_rank + data_rank ;
  int* an  = al  + data_rank ;
  int* as  = an  + data_rank ;
  int* tn  = as  + data_rank ;
  int* b   = tn  + data_rank ;
  int* fmt = b   + data_rank ;
  int* d   = fmt + data_rank ;
  int* ep  = d   + data_rank ;
  int* ed  = ep  + embed_rank ;

  data = (char*) _data ;

  //{
  //  for(int i = 0 ; i < data_rank ; i++)
  //    printf("Cludge :: Cludge: l [%d] = %d n [%d] = %d s [%d] = %d "
  //           "al [%d] = %d an [%d] = %d as [%d] = %d "
  //           "tn [%d] = %d b [%d] = %d fmt [%d] = %d d [%d] = %d\n",
  //           i, l [i], i, n [i], i, s [i],
  //           i, al [i], i, an [i], i, as [i],
  //           i, tn [i], i, b [i], i, fmt [i], i, d [i]) ;
  //  for(int j = 0 ; j < embed_rank ; j++)
  //    printf("Cludge :: Cludge: ep [%d] = %d ed [%d] = %d\n",
  //           j, ep [j], j, ed [j]) ;
  //}  // debug
    
  int size = typeLen ;
  int i_frm = 0 ;
  for(int i_dat = 0 ; i_dat < data_rank ; i_dat++) {
    Range *rng_prm ;
    if(d [i_dat] == -1)
      rng_prm = new Range(tn [i_dat]) ;  // Collapsed range.
    else {
      //printf("d = %d, dim(d) = %x\n",
      //       d [i_dat], procGrp->procs->dim(d [i_dat] + 1)) ;  // debug

      rng_prm = new Range(tn [i_dat],
                          procGrp->procs->dim(d [i_dat] + 1), cyc, b [i_dat]) ;
    }

    Range rng_dat(an [i_dat], al [i_dat], as [i_dat], *rng_prm) ;

    Fudge* str = new Fudge ;
    size = str->init(&rng_dat, size) ;

    if(n [i_dat] == -1) {
      Embed e(l [i_dat], rng_dat) ;
      if(rng_dat.distributed())
        procGrp->deactivate(&e) ;
      data += str->offset(&e) ;

      delete str ;
    }
    else {
      strs [i_frm] = str ;
      if(n [i_dat] == -2)
        rngs [i_frm] = new Range(rng_dat) ;
      else
        rngs [i_frm] = new Range(n [i_dat], l [i_dat], s [i_dat], rng_dat) ;
      i_frm++ ;
    }

    delete rng_prm ;
  }

  rank = i_frm ;

  for(int i_emb = 0 ; i_emb < embed_rank ; i_emb++) {
    Dimension* dim = procGrp->procs->dim(ed [i_emb] + 1) ;
    if(dim)
      procGrp->deactivate(dim, ep [i_emb]) ;
  }
}


Cludge :: ~Cludge() {
  for(int i = rank - 1 ; i >= 0 ; i--) {
    delete rngs [i] ;
    delete strs [i] ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Fudge': Associated variant of `Stride' structure.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int Fudge :: init(Range* x, const int dataLen) {

  // The representation of the stride structure is independent
  // of the `x' coordinate!

  const int stride    = x->stride ;
  const int layerSize = x->layerSize ;
  const int s_lo      = x->base   ;
  const int s_hi      = x->base + (x->extent - 1) * stride ;

  int blockWidth ;
  int top_v ;

  if(stride == 0) AdlibError("Fudge :: init: stride is zero") ;

  if(stride > 0) {
    blockWidth = (x->blockSize + stride - 1) / stride ;
    top_v      = s_hi / layerSize ;

    stride_v = blockWidth * dataLen ;
    base_v   = 0 ;           // Note the difference!
    stride_h = dataLen ;
    norm_h   = stride ;
  }
  else {
    blockWidth = (x->blockSize - stride - 1) / (-stride) ;
    top_v      = s_lo / layerSize ;

    stride_v = blockWidth * dataLen ;
    base_v   = 0 ;           // Note the difference!
    stride_h = dataLen ;
    norm_h   = -stride ;
  }

  stodge = !x->distributed() ;  // bodge

  return (top_v - base_v + 1) * blockWidth * dataLen ;
}


